import sys
import click

from vantage6.client import UserClient
from vantage6.cli.utils import error
from vantage6.cli.test.common.diagnostic_runner import DiagnosticRunner


@click.command()
@click.option("--host", type=str, default="http://localhost",
              help="URL of the server")
@click.option("--port", type=int, default=5000, help="Port of the server")
@click.option("--api-path", type=str, default="/api",
              help="API path of the server")
@click.option("--username", type=str, default="root",
              help="Username of vantage6 user account to create the task with")
@click.option("--password", type=str, default="root",
              help="Password of vantage6 user account to create the task with")
@click.option("--collaboration", type=int, default=1,
              help="ID of the collaboration to create the task in")
@click.option("-o", "--organizations", type=int, default=[], multiple=True,
              help="ID(s) of the organization(s) to create the task for")
@click.option("--all-nodes", is_flag=True,
              help="Run the diagnostic test on all nodes in the collaboration")
@click.option("--online-only", is_flag=True,
              help="Run the diagnostic test on only nodes that are online")
@click.option("--no-vpn", is_flag=True,
              help="Don't execute VPN tests")
def cli_test_features(
    host: str, port: int, api_path: str, username: str, password: str,
    collaboration: int, organizations: list[int] | None, all_nodes: bool,
    online_only: bool, no_vpn: bool
) -> list[dict]:
    """
    Run diagnostic checks on an existing vantage6 network.

    This command will create a task in the requested collaboration that will
    test the functionality of vantage6, and will report back the results.
    """
    if all_nodes and organizations:
        error("Cannot use --all-nodes and --organization at the same time.")
        sys.exit(1)

    if all_nodes or not organizations:
        organizations = None

    client = UserClient(host=host, port=port, path=api_path,
                        log_level='critical')
    client.authenticate(username=username, password=password)
    client.setup_encryption(None)
    diagnose = DiagnosticRunner(client, collaboration, organizations,
                                online_only)
    res = diagnose(base=True, vpn=not no_vpn)
    return res



import subprocess

import click
import docker
from colorama import (Fore, Style)

from vantage6.common import info, error, debug as debug_msg
from vantage6.common.docker.addons import check_docker_running
from vantage6.common.globals import APPNAME
from vantage6.cli.context import ServerContext
from vantage6.cli.server.common import click_insert_context


@click.command()
@click_insert_context
def cli_server_shell(ctx: ServerContext) -> None:
    """
    Run an iPython shell within a running server. This can be used to modify
    the database.

    NOTE: using the shell is no longer recommended as there is no validation on
    the changes that you make. It is better to use the Python client or a
    graphical user interface instead.
    """
    # will print an error if not
    check_docker_running()

    docker_client = docker.from_env()

    running_servers = docker_client.containers.list(
        filters={"label": f"{APPNAME}-type=server"})

    if ctx.docker_container_name not in [s.name for s in running_servers]:
        error(f"Server {Fore.RED}{ctx.name}{Style.RESET_ALL} is not running?")
        return

    try:
        subprocess.run(['docker', 'exec', '-it', ctx.docker_container_name,
                        'vserver-local', 'shell', '-c', '/mnt/config.yaml'])
    except Exception as e:
        info("Failed to start subprocess...")
        debug_msg(e)



from pathlib import Path
import csv

import click
from jinja2 import Environment, FileSystemLoader
from colorama import Fore, Style

from vantage6.common.globals import APPNAME
from vantage6.common import info, error, generate_apikey

from vantage6.cli.globals import PACKAGE_FOLDER
from vantage6.cli.context import ServerContext, NodeContext
from vantage6.cli.server.common import get_server_context
from vantage6.cli.server.import_ import cli_server_import
from vantage6.cli.utils import prompt_config_name


def create_dummy_data(node_name: str, dev_folder: Path) -> Path:
    """Synthesize csv dataset.

    Parameters
    ----------
    node_name : str
        Name of node to be used as part of dataset.
    dev_folder : Path
        Path to the dev folder.

    Returns
    -------
    Path
        Directory the data is saved in.
    """
    header = ['name', 'mask', 'weapon', 'age']
    data = [
        ['Raphael', 'red', 'sai', 44],
        ['Donatello', 'purple', 'bo staff', 60],
    ]

    data_file = dev_folder / f"df_{node_name}.csv"
    with open(data_file, 'w', encoding='UTF8', newline='') as f:
        writer = csv.writer(f)

        # write the header
        writer.writerow(header)

        # write the data
        for row in data:
            writer.writerow(row)

        f.close()

    info(f"Spawned dataset for {Fore.GREEN}{node_name}{Style.RESET_ALL}, "
         f"writing to {Fore.GREEN}{data_file}{Style.RESET_ALL}")
    return data_file


def create_node_config_file(server_url: str, port: int, config: dict,
                            server_name: str) -> None:
    """Create a node configuration file (YAML).

    Creates a node configuration for a simulated organization. Organization ID
    is used for generating both the organization name and node_name as each
    organization only houses one node.

    Parameters
    ----------
    server_url : str
        Url of the dummy server.
    port : int
        Port of the dummy server.
    config : dict
        Configuration dictionary containing org_id, api_key, node name and
        additional user_defined_config.
    server_name : str
        Configuration name of the dummy server.
    """
    environment = Environment(
        loader=FileSystemLoader(PACKAGE_FOLDER / APPNAME / "cli" / "template"),
        trim_blocks=True, lstrip_blocks=True, autoescape=True)
    template = environment.get_template("node_config.j2")

    # TODO: make this name specific to the server it connects
    node_name = config['node_name']
    folders = NodeContext.instance_folders('node', node_name, False)
    path_to_dev_dir = Path(folders['dev'] / server_name)
    path_to_dev_dir.mkdir(parents=True, exist_ok=True)
    dummy_datafile = create_dummy_data(node_name, path_to_dev_dir)

    path_to_data_dir = Path(folders['data'])
    path_to_data_dir.mkdir(parents=True, exist_ok=True)
    full_path = Path(folders['config'] / f'{node_name}.yaml')

    if full_path.exists():
        error(f"Node configuration file already exists: {full_path}")
        exit(1)

    node_config = template.render({
        "api_key": config['api_key'],
        "databases": {
            "default": dummy_datafile
        },
        "logging": {
            "file": f'{node_name}.log'
        },
        "port": port,
        "server_url": server_url,
        "task_dir": str(path_to_data_dir),
        "user_provided_config": config['user_defined_config']
    })

    try:
        with open(full_path, 'x') as f:
            f.write(node_config)
    except Exception as e:
        error(f"Could not write node configuration file: {e}")
        exit(1)

    info(f"Spawned node for organization {Fore.GREEN}{config['org_id']}"
         f"{Style.RESET_ALL}")


def _read_extra_config_file(extra_config_file: Path | None) -> str:
    """Reads extra configuration file.

    Parameters
    ----------
    extra_config_file : Path | None
        Path to file with additional configuration.

    Returns
    -------
    str
        Extra configuration file content
    """
    if extra_config_file:
        # read the YAML file as string, so it can be appended to the
        # configuration easily
        with open(extra_config_file, 'r', encoding='utf-8') as f:
            return f.read()
    return ''


def generate_node_configs(
    num_nodes: int, server_url: str, port: int, server_name: str,
    extra_node_config: Path | None
) -> list[dict]:
    """Generates ``num_nodes`` node configuration files.

    Parameters
    ----------
    num_nodes : int
        Integer to determine how many configurations to create.
    server_url : str
        Url of the dummy server.
    port : int
        Port of the dummy server.
    server_name : str
        Configuration name of the dummy server.
    extra_node_config : Path | None
        Path to file with additional node configuration.

    Returns
    -------
    list[dict]
        List of dictionaries containing node configurations.
    """
    configs = []
    extra_config = _read_extra_config_file(extra_node_config)
    for i in range(num_nodes):
        config = {
            'org_id': i + 1,
            'api_key': generate_apikey(),
            'node_name': f"{server_name}_node_{i + 1}",
            "user_defined_config": extra_config
        }
        create_node_config_file(server_url, port, config, server_name)
        configs.append(config)

    return configs


def create_vserver_import_config(node_configs: list[dict], server_name: str) \
        -> Path:
    """Create server configuration import file (YAML).

    Utilized by the ``v6 server import`` command.

    Parameters
    ----------
    node_configs : list[dict]
        List of dictionaries containing the node configurations, returned from
        ``generate_node_configs()``.
    server_name : str
        Server name.

    Returns
    -------
    Path
        Path object where the server import configuration is stored.
    """
    environment = Environment(
        loader=FileSystemLoader(PACKAGE_FOLDER / APPNAME / "cli" / "template"),
        trim_blocks=True, lstrip_blocks=True, autoescape=True)
    template = environment.get_template("server_import_config.j2")

    organizations = []
    collaboration = {'name': 'demo', 'participants': []}
    for config in node_configs:
        org_id = config['org_id']
        org_data = {'name': f"org_{org_id}"}

        organizations.append(org_data)
        collaboration['participants'].append({'name': f"org_{org_id}",
                                              'api_key': config['api_key']})
    organizations[0]['make_admin'] = True
    info(f"Organization {Fore.GREEN}{node_configs[0]['org_id']}"
         f"{Style.RESET_ALL} is the admin")

    server_import_config = template.render(organizations=organizations,
                                           collaboration=collaboration)
    folders = ServerContext.instance_folders("server", server_name, False)

    demo_dir = Path(folders['dev'])
    demo_dir.mkdir(parents=True, exist_ok=True)
    full_path = demo_dir / f'{server_name}.yaml'
    if full_path.exists():
        error(f"Server configuration file already exists: {full_path}")
        exit(1)

    try:
        with open(full_path, 'x') as f:
            f.write(server_import_config)
            info("Server import configuration ready, writing to "
                 f"{Fore.GREEN}{full_path}{Style.RESET_ALL}")
    except Exception as e:
        error(f"Could not write server import configuration file: {e}")
        exit(1)

    return full_path


def create_vserver_config(
    server_name: str, port: int, extra_config_file: Path
) -> Path:
    """Creates server configuration file (YAML).

    Parameters
    ----------
    server_name : str
        Server name.
    port : int
        Server port.
    extra_config_file : Path
        Path to file with additional server configuration.

    Returns
    -------
    Path
        Path object where server configuration is stored.
    """
    environment = Environment(
        loader=FileSystemLoader(PACKAGE_FOLDER / APPNAME / "cli" / "template"),
        trim_blocks=True, lstrip_blocks=True, autoescape=True)

    extra_config = _read_extra_config_file(extra_config_file)

    template = environment.get_template("server_config.j2")
    server_config = template.render(
        port=port,
        jwt_secret_key=generate_apikey(),
        user_provided_config=extra_config
    )
    folders = ServerContext.instance_folders(
        instance_type='server', instance_name=server_name,
        system_folders=True)

    config_dir = Path(folders['config'] / server_name)
    config_dir.mkdir(parents=True, exist_ok=True)
    full_path = folders["config"] / f'{server_name}.yaml'
    if full_path.exists():
        error(f"Server configuration file already exists: {full_path}")
        exit(1)

    try:
        with open(full_path, 'x') as f:
            f.write(server_config)
            info("Server configuration read, writing to "
                 f"{Fore.GREEN}{full_path}{Style.RESET_ALL}")
    except Exception as e:
        error(f"Could not write server configuration file: {e}")
        exit(1)

    return full_path


def demo_network(
    num_nodes: int, server_url: str, server_port: int, server_name: str,
    extra_server_config: Path, extra_node_config: Path
) -> tuple[list[dict], Path, Path]:
    """Generates the demo network.

    Parameters
    ----------
    num_nodes : int
        Integer to determine how many configurations to create.
    server_url : str
        Url of the dummy server.
    server_port : int
        Port of the dummy server.
    server_name : str
        Server name.
    extra_server_config : Path
        Path to file with additional server configuration.
    extra_node_config : Path
        Path to file with additional node configuration.

    Returns
    -------
    tuple[list[dict], Path, Path]
        Tuple containing node, server import and server configurations.
    """
    node_configs = generate_node_configs(num_nodes, server_url, server_port,
                                         server_name, extra_node_config)
    server_import_config = create_vserver_import_config(node_configs,
                                                        server_name)
    server_config = create_vserver_config(server_name, server_port,
                                          extra_server_config)
    return (node_configs, server_import_config, server_config)


@click.command()
@click.option('-n', '--name', default=None, type=str,
              help="Name for your development setup")
@click.option('--num-nodes', type=int, default=3,
              help='Generate this number of nodes in the development network')
@click.option('--server-url', type=str, default='http://host.docker.internal',
              help='Server URL to point to. If you are using Docker Desktop, '
              'the default http://host.docker.internal should not be changed.')
@click.option('-p', '--server-port', type=int, default=5000,
              help='Port to run the server on. Default is 5000.')
@click.option('-i', '--image', type=str, default=None,
              help='Server docker image to use when setting up resources for '
              'the development server')
@click.option('--extra-server-config', type=click.Path(exists=True),
              default=None, help='YAML File with additional server '
              'configuration. This will be appended to the server '
              'configuration file')
@click.option('--extra-node-config', type=click.Path('rb'), default=None,
              help='YAML File with additional node configuration. This will be'
              ' appended to each of the node configuration files')
@click.pass_context
def create_demo_network(
    click_ctx: click.Context, name: str, num_nodes: int, server_url: str,
    server_port: int, image: str = None, extra_server_config: Path = None,
    extra_node_config: Path = None
) -> dict:
    """Creates a demo network.

    Creates server instance as well as its import configuration file. Server
    name is set to 'dev_default_server'. Generates `n` node configurations, but
    by default this is set to 3. Then runs a Batch import of
    organizations/collaborations/users and tasks.
    """
    server_name = prompt_config_name(name)
    if not ServerContext.config_exists(server_name):
        demo = demo_network(
            num_nodes, server_url, server_port, server_name,
            extra_server_config, extra_node_config
        )
        info(f"Created {Fore.GREEN}{len(demo[0])}{Style.RESET_ALL} node "
             f"configuration(s), attaching them to {Fore.GREEN}{server_name}"
             f"{Style.RESET_ALL}.")
    else:
        error(f"Configuration {Fore.RED}{server_name}{Style.RESET_ALL} "
              "already exists!")
        exit(1)
    (node_config, server_import_config, server_config) = demo
    ctx = get_server_context(server_name, True)
    click_ctx.invoke(
        cli_server_import, ctx=ctx, file=server_import_config, drop_all=False,
        image=image, mount_src='', keep=False, wait=True
    )
    return {
        "node_configs": node_config,
        "server_import_config": server_import_config,
        "server_config": server_config
    }



import os
from pathlib import Path

import click
import questionary as q
from copier import run_update

from vantage6.cli.utils import info


@click.command()
@click.option('-d', '--dir', 'directory', default=None, type=str,
              help="Directory to put the algorithm into")
def cli_algorithm_update(directory: str) -> dict:
    """Update an algorithm template

    When a new version of the algorithm template is released, you can update
    your existing algorithm template (which you created with
    `v6 algorithm create`) by running this command.
    """
    if not directory:
        default_dir = str(Path(os.getcwd()))
        directory = q.text(
            "Algorithm directory:",
            default=default_dir
        ).ask()

    run_update(directory, overwrite=True)

    info("Template updated!")



"""RabbitMQ utilities."""


def split_rabbitmq_uri(rabbit_uri: str) -> dict:
    """
    Get details (user, pass, host, vhost, port) from a RabbitMQ uri.

    Parameters
    ----------
    rabbit_uri: str
        URI of RabbitMQ service ('amqp://$user:$pass@$host:$port/$vhost')

    Returns
    -------
    dict[str]
        The vhost defined in the RabbitMQ URI
    """
    (user_details, location_details) = rabbit_uri.split('@', 1)
    (user, password) = user_details.split('/')[-1].split(':', 1)
    (host, remainder) = location_details.split(':', 1)
    port, vhost = remainder.split('/', 1)
    return {
        'user': user,
        'password': password,
        'host': host,
        'port': port,
        'vhost': vhost,
    }



import click
from colorama import (Fore, Style)

from vantage6.common import info, error, check_config_writeable
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS
from vantage6.cli.context import ServerContext
from vantage6.cli.configuration_wizard import configuration_wizard
from vantage6.cli.utils import check_config_name_allowed, prompt_config_name


@click.command()
@click.option('-n', '--name', default=None,
              help="name of the configuration you want to use.")
@click.option('--system', 'system_folders', flag_value=True)
@click.option('--user', 'system_folders', flag_value=False,
              default=DEFAULT_SERVER_SYSTEM_FOLDERS)
def cli_server_new(name: str, system_folders: bool) -> None:
    """
    Create a new server configuration.
    """
    name = prompt_config_name(name)

    # check if name is allowed for docker volume, else exit
    check_config_name_allowed(name)

    # check that this config does not exist
    try:
        if ServerContext.config_exists(name, system_folders):
            error(f"Configuration {Fore.RED}{name}{Style.RESET_ALL} already "
                  "exists!")
            exit(1)
    except Exception as e:
        error(e)
        exit(1)

    # Check that we can write in this folder
    if not check_config_writeable(system_folders):
        error("Your user does not have write access to all folders. Exiting")
        info(f"Create a new server using '{Fore.GREEN}v6 server new "
             f"--user{Style.RESET_ALL}' instead!")
        exit(1)

    # create config in ctx location
    cfg_file = configuration_wizard("server", name, system_folders)
    info(f"New configuration created: {Fore.GREEN}{cfg_file}{Style.RESET_ALL}")

    # info(f"root user created.")
    flag = "" if system_folders else "--user"
    info(f"You can start the server by running {Fore.GREEN}v6 server start "
         f"{flag}{Style.RESET_ALL}")



import click

from vantage6.cli.context import ServerContext, NodeContext
from vantage6.cli.server.common import click_insert_context
from vantage6.cli.server.stop import cli_server_stop
from vantage6.cli.node.stop import cli_node_stop


@click.command()
@click_insert_context
@click.pass_context
def stop_demo_network(click_ctx: click.Context, ctx: ServerContext) -> None:
    """ Stops a demo network's server and nodes.

    Select a server configuration to stop that server and the nodes attached
    to it.
    """
    # stop the server
    click_ctx.invoke(
        cli_server_stop, name=ctx.name, system_folders=True, all_servers=False
    )

    # stop the nodes
    configs, _ = NodeContext.available_configurations(False)
    node_names = [
        config.name for config in configs if f'{ctx.name}_node_' in config.name
    ]
    for name in node_names:
        click_ctx.invoke(cli_node_stop, name=name, system_folders=False,
                         all_nodes=False, force=False)



import click
import docker
from colorama import Fore, Style

from vantage6.common import warning
from vantage6.common.globals import APPNAME
from vantage6.common.docker.addons import check_docker_running
from vantage6.cli.context import NodeContext
from vantage6.cli.node.common import find_running_node_names


@click.command()
def cli_node_list() -> None:
    """
    Lists all node configurations.

    Note that this command cannot find node configuration files in custom
    directories.
    """

    check_docker_running()
    client = docker.from_env()

    running_node_names = find_running_node_names(client)

    header = \
        "\nName"+(21*" ") + \
        "Status"+(10*" ") + \
        "System/User"

    click.echo(header)
    click.echo("-"*len(header))

    running = Fore.GREEN + "Running" + Style.RESET_ALL
    stopped = Fore.RED + "Not running" + Style.RESET_ALL

    # system folders
    configs, f1 = NodeContext.available_configurations(
        system_folders=True)
    for config in configs:
        status = running if f"{APPNAME}-{config.name}-system" in \
            running_node_names else stopped
        click.echo(
            f"{config.name:25}"
            f"{status:25}System "
        )

    # user folders
    configs, f2 = NodeContext.available_configurations(
        system_folders=False)
    for config in configs:
        status = running if f"{APPNAME}-{config.name}-user" in \
            running_node_names else stopped
        click.echo(
            f"{config.name:25}"
            f"{status:25}User   "
        )

    click.echo("-"*53)
    if len(f1)+len(f2):
        warning(
             f"{Fore.RED}Failed imports: {len(f1)+len(f2)}{Style.RESET_ALL}")



import click
import docker

from colorama import (Fore, Style)

from vantage6.common import warning
from vantage6.common.docker.addons import check_docker_running
from vantage6.common.globals import APPNAME
from vantage6.cli.context import ServerContext


@click.command()
def cli_server_configuration_list() -> None:
    """
    Print the available server configurations.
    """
    check_docker_running()
    client = docker.from_env()

    running_server = client.containers.list(
        filters={"label": f"{APPNAME}-type=server"})
    running_node_names = []
    for node in running_server:
        running_node_names.append(node.name)

    header = \
        "\nName"+(21*" ") + \
        "Status"+(10*" ") + \
        "System/User"

    click.echo(header)
    click.echo("-"*len(header))

    running = Fore.GREEN + "Running" + Style.RESET_ALL
    stopped = Fore.RED + "Not running" + Style.RESET_ALL

    # system folders
    configs, f1 = ServerContext.available_configurations(system_folders=True)
    for config in configs:
        status = running if f"{APPNAME}-{config.name}-system-server" in \
            running_node_names else stopped
        click.echo(
            f"{config.name:25}"
            f"{status:25} System "
        )

    # user folders
    configs, f2 = ServerContext.available_configurations(system_folders=False)
    for config in configs:
        status = running if f"{APPNAME}-{config.name}-user-server" in \
            running_node_names else stopped
        click.echo(
            f"{config.name:25}"
            f"{status:25} User   "
        )

    click.echo("-"*85)
    if len(f1)+len(f2):
        warning(
             f"{Fore.RED}Failed imports: {len(f1)+len(f2)}{Style.RESET_ALL}")



from typing import Iterable
from functools import wraps

from docker.client import DockerClient
import click
from colorama import Fore, Style

from vantage6.common import error, info
from vantage6.common.globals import STRING_ENCODING, APPNAME
from vantage6.common.docker.addons import remove_container, get_container

from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS
from vantage6.cli.context import ServerContext
from vantage6.cli.configuration_wizard import select_configuration_questionaire


def click_insert_context(func: callable) -> callable:
    """
    Supply the Click function with additional context parameters. The context
    is then passed to the function as the first argument.

    Parameters
    ----------
    func : Callable
        function you want the context to be passed to

    Returns
    -------
    Callable
        Click function with context
    """
    @click.option('-n', '--name', default=None,
                  help="Name of the configuration you want to use.")
    @click.option('-c', '--config', default=None,
                  help='Absolute path to configuration-file; overrides NAME')
    @click.option('--system', 'system_folders', flag_value=True,
                  help='Use system folders instead of user folders. This is '
                  'the default')
    @click.option('--user', 'system_folders', flag_value=False,
                  default=DEFAULT_SERVER_SYSTEM_FOLDERS,
                  help='Use user folders instead of system folders')
    @wraps(func)
    def func_with_context(name: str, config: str, system_folders: bool, *args,
                          **kwargs) -> callable:
        """
        Decorator function that adds the context to the function.

        Returns
        -------
        Callable
            Decorated function
        """
        # path to configuration file always overrides name
        if config:
            ctx = ServerContext.from_external_config_file(
                config,
                system_folders
            )
        elif 'ctx' in kwargs:
            # if ctx is already in kwargs (typically when one click command
            # calls another internally), use that existing ctx
            ctx = kwargs.pop('ctx')
        else:
            # in case no name, ctx or config file is supplied, ask the user to
            # select an existing config by name
            if not name:
                try:
                    # select configuration if none supplied
                    name = select_configuration_questionaire(
                        "server", system_folders
                    )
                except Exception:
                    error("No configurations could be found!")
                    exit(1)

            ctx = get_server_context(name, system_folders)
        return func(ctx, *args, **kwargs)

    return func_with_context


def get_server_context(name: str, system_folders: bool) \
        -> ServerContext:
    """
    Load the server context from the configuration file.

    Parameters
    ----------
    name : str
        Name of the server to inspect
    system_folders : bool
        Wether to use system folders or if False, the user folders

    Returns
    -------
    ServerContext
        Server context object
    """
    if not ServerContext.config_exists(name, system_folders):
        scope = "system" if system_folders else "user"
        error(
            f"Configuration {Fore.RED}{name}{Style.RESET_ALL} does not "
            f"exist in the {Fore.RED}{scope}{Style.RESET_ALL} folders!"
        )
        exit(1)

    # We do not want to log this here, we do this in the container and not on
    # the host. We only want CLI logging here.
    ServerContext.LOGGING_ENABLED = False

    # create server context, and initialize db
    ctx = ServerContext(name, system_folders=system_folders)

    return ctx


def print_log_worker(logs_stream: Iterable[bytes]) -> None:
    """
    Print the logs from the docker container to the terminal.

    Parameters
    ----------
    logs_stream : Iterable[bytes]
        Output of the `container.attach(.)` method
    """
    for log in logs_stream:
        print(log.decode(STRING_ENCODING), end="")


def stop_ui(client: DockerClient, ctx: ServerContext) -> None:
    """
    Check if the UI container is running, and if so, stop and remove it.

    Parameters
    ----------
    client : DockerClient
        Docker client
    ctx : ServerContext
        Server context object
    """
    ui_container_name = f"{APPNAME}-{ctx.name}-{ctx.scope}-ui"
    ui_container = get_container(client, name=ui_container_name)
    if ui_container:
        remove_container(ui_container, kill=True)
        info(f"Stopped the {Fore.GREEN}{ui_container_name}"
             f"{Style.RESET_ALL} User Interface container.")



import subprocess
import itertools
from shutil import rmtree
from pathlib import Path

import click

from vantage6.common import info
from vantage6.cli.context import ServerContext, NodeContext
from vantage6.cli.server.common import click_insert_context
from vantage6.cli.server.remove import cli_server_remove
from vantage6.cli.utils import remove_file


@click.command()
@click_insert_context
@click.pass_context
def remove_demo_network(
    click_ctx: click.Context, ctx: ServerContext
) -> None:
    """ Remove all related demo network files and folders.

    Select a server configuration to remove that server and the nodes attached
    to it.
    """

    # remove the server
    for handler in itertools.chain(ctx.log.handlers, ctx.log.root.handlers):
        handler.close()
    click_ctx.invoke(cli_server_remove, ctx=ctx, force=True)

    # removing the server import config
    info("Deleting demo import config file")
    server_configs = ServerContext.instance_folders("server", ctx.name,
                                                    system_folders=False)
    import_config_to_del = Path(server_configs['dev']) / f"{ctx.name}.yaml"
    remove_file(import_config_to_del, 'import_configuration')

    # also remove the server folder
    server_configs = ServerContext.instance_folders("server", ctx.name,
                                                    system_folders=True)
    server_folder = server_configs['data']
    if server_folder.is_dir():
        rmtree(server_folder)

    # remove the nodes
    configs, _ = NodeContext.available_configurations(system_folders=False)
    node_names = [
        config.name for config in configs if f'{ctx.name}_node_' in config.name
    ]
    for name in node_names:
        node_ctx = NodeContext(name, False)
        for handler in itertools.chain(node_ctx.log.handlers,
                                       node_ctx.log.root.handlers):
            handler.close()
        subprocess.run(
            ["v6", "node", "remove", "-n", name, "--user", "--force"]
        )

    # remove data files attached to the network
    data_dirs_nodes = NodeContext.instance_folders('node', '', False)['dev']
    rmtree(Path(data_dirs_nodes / ctx.name))



"""
This module contains global variables that are used throughout the CLI.
"""
from pathlib import Path
from vantage6.common.globals import APPNAME

#
#   SERVER SETTINGS
#
DEFAULT_SERVER_SYSTEM_FOLDERS = True

#
#   NODE SETTINGS
#
DEFAULT_NODE_SYSTEM_FOLDERS = False

#
#   INSTALLATION SETTINGS
#
PACKAGE_FOLDER = Path(__file__).parent.parent.parent

# FIXME BvB 22-06-28 think this is also defined in the node globals, and this
# one appears not to be used
NODE_PROXY_SERVER_HOSTNAME = "proxyserver"

DATA_FOLDER = PACKAGE_FOLDER / APPNAME / "_data"

# with open(Path(PACKAGE_FOLDER) / APPNAME / "cli" / "VERSION") as f:
#     VERSION = f.read()

# Maximum time to start up RabbitMQ in seconds
RABBIT_TIMEOUT = 300

# Default port to start the UI on
DEFAULT_UI_PORT = 5001

# Location of repository to create new algorithm templates from
ALGORITHM_TEMPLATE_REPO = "gh:vantage6/v6-algorithm-template.git"

# image to use for diagnostics in `v6 test` commands
DIAGNOSTICS_IMAGE = "harbor2.vantage6.ai/algorithms/diagnostic"



import unittest

from unittest.mock import MagicMock, patch
from pathlib import Path
from click.testing import CliRunner

from vantage6.cli.globals import APPNAME
from vantage6.cli.server.start import cli_server_start
from vantage6.cli.server.list import cli_server_configuration_list
from vantage6.cli.server.files import cli_server_files
from vantage6.cli.server.import_ import cli_server_import
from vantage6.cli.server.new import cli_server_new
from vantage6.cli.server.stop import cli_server_stop
from vantage6.cli.server.attach import cli_server_attach


class ServerCLITest(unittest.TestCase):

    @patch("vantage6.cli.server.start.NetworkManager")
    @patch("vantage6.cli.server.start.docker.types.Mount")
    @patch("os.makedirs")
    @patch("vantage6.cli.server.start.pull_if_newer")
    @patch("vantage6.cli.server.common.ServerContext")
    @patch("vantage6.cli.server.start.docker.from_env")
    @patch("vantage6.cli.server.start.check_docker_running", return_value=True)
    def test_start(self, docker_check, containers, context,
                   pull, os_makedirs, mount, network_manager):
        """Start server without errors"""
        container1 = MagicMock()
        container1.containers.name = f"{APPNAME}-iknl-system"
        containers.containers.list.return_value = [container1]
        containers.containers.run.return_value = True

        # mount.types.Mount.return_value = MagicMock()

        ctx = MagicMock(
            config={
                'uri': 'sqlite:///file.db',
                'port': 9999
            },
            config_file="/config.yaml",
            data_dir=Path(".")
        )
        ctx.config_exists.return_value = True
        ctx.name = 'not-running'
        context.return_value = ctx

        runner = CliRunner()
        result = runner.invoke(cli_server_start, ["--name", "not-running"])

        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.server.common.ServerContext")
    @patch("docker.DockerClient.containers")
    @patch("vantage6.cli.server.list.check_docker_running", return_value=True)
    def test_configuration_list(self, docker_check, containers, context):
        """Configuration list without errors."""
        container1 = MagicMock()
        container1.name = f"{APPNAME}-iknl-system"
        containers.list.return_value = [container1]

        config = MagicMock()
        config.name = "iknl"
        context.available_configurations.return_value = ([config], [])

        runner = CliRunner()
        result = runner.invoke(cli_server_configuration_list)

        self.assertEqual(result.exit_code, 0)
        self.assertIsNone(result.exception)

    @patch("vantage6.cli.server.common.ServerContext")
    def test_files(self, context):
        """Configuration files without errors."""

        ctx = context.return_value = MagicMock(
            log_file="/log_file.log",
            config_file="/iknl.yaml"
        )
        ctx.get_database_uri.return_value = "sqlite:///test.db"

        runner = CliRunner()
        result = runner.invoke(cli_server_files, ["--name", "iknl"])

        self.assertIsNone(result.exception)
        self.assertEqual(result.exit_code, 0)

    @patch("docker.DockerClient.containers")
    @patch("vantage6.cli.server.import_.print_log_worker")
    @patch("vantage6.cli.server.import_.click.Path")
    @patch("vantage6.cli.server.import_.check_docker_running",
           return_value=True)
    @patch("vantage6.cli.server.common.ServerContext")
    def test_import(self, context, docker_check, click_path, log, containers):
        """Import entities without errors."""
        click_path.return_value = MagicMock()

        ctx = MagicMock()
        ctx.name = 'some-name'
        context.return_value = ctx

        runner = CliRunner()
        with runner.isolated_filesystem():
            with open("some.yaml", "w") as fp:
                fp.write("does-not-matter")
            result = runner.invoke(cli_server_import, [
                "--name", "iknl", "some.yaml"
            ])

        self.assertIsNone(result.exception)
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.server.new.configuration_wizard")
    @patch("vantage6.cli.server.new.check_config_writeable")
    @patch("vantage6.cli.server.new.ServerContext")
    def test_new(self, context, permissions, wizard):
        """New configuration without errors."""

        context.config_exists.return_value = False
        permissions.return_value = True
        wizard.return_value = "/some/file.yaml"

        runner = CliRunner()
        result = runner.invoke(cli_server_new, ["--name", "iknl"])

        self.assertIsNone(result.exception)
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.server.stop.docker.from_env")
    def test_stop(self, containers):
        """Stop server without errors."""

        container1 = MagicMock()
        container1.name = f"{APPNAME}-iknl-system-server"
        containers.containers.list.return_value = [container1]

        runner = CliRunner()
        result = runner.invoke(cli_server_stop, ["--name", "iknl"])

        self.assertIsNone(result.exception)
        self.assertEqual(result.exit_code, 0)

    @patch("vantage6.cli.server.attach.time.sleep")
    @patch("docker.DockerClient.containers")
    def test_attach(self, containers, sleep):
        """Attach log to the console without errors."""
        container1 = MagicMock()
        container1.name = f"{APPNAME}-iknl-system-server"
        containers.list.return_value = [container1]

        sleep.side_effect = KeyboardInterrupt("Boom!")

        runner = CliRunner()
        result = runner.invoke(cli_server_attach, ["--name", "iknl"])

        self.assertIsNone(result.exception)
        self.assertEqual(result.exit_code, 0)



import os
from threading import Thread
import time

import click
import docker
from colorama import (Fore, Style)
from sqlalchemy.engine.url import make_url
from docker.client import DockerClient

from vantage6.common import info, warning, error
from vantage6.common.docker.addons import (
    pull_if_newer, check_docker_running
)
from vantage6.common.docker.network_manager import NetworkManager
from vantage6.common.globals import (
    APPNAME,
    DEFAULT_DOCKER_REGISTRY,
    DEFAULT_SERVER_IMAGE,
    DEFAULT_UI_IMAGE
)

from vantage6.cli.globals import DEFAULT_UI_PORT
from vantage6.cli.context import ServerContext
from vantage6.cli.utils import check_config_name_allowed
from vantage6.cli.rabbitmq.queue_manager import RabbitMQManager
from vantage6.cli.server.common import (
    click_insert_context, print_log_worker, stop_ui
)


@click.command()
@click.option('--ip', default=None, help='IP address to listen on')
@click.option('-p', '--port', default=None, type=int, help='Port to listen on')
@click.option('-i', '--image', default=None, help="Server Docker image to use")
@click.option('--with-ui', 'start_ui', flag_value=True, default=False,
              help="Start the graphical User Interface as well")
@click.option('--ui-port', default=None, type=int,
              help="Port to listen on for the User Interface")
@click.option('--with-rabbitmq', 'start_rabbitmq', flag_value=True,
              default=False, help="Start RabbitMQ message broker as local "
              "container - use in development only")
@click.option('--rabbitmq-image', default=None,
              help="RabbitMQ docker image to use")
@click.option('--keep/--auto-remove', default=False,
              help="Keep image after server has stopped. Useful for debugging")
@click.option('--mount-src', default='',
              help="Override vantage6 source code in container with the source"
              " code in this path")
@click.option('--attach/--detach', default=False,
              help="Print server logs to the console after start")
@click_insert_context
def cli_server_start(ctx: ServerContext, ip: str, port: int, image: str,
                     start_ui: bool, ui_port: int, start_rabbitmq: bool,
                     rabbitmq_image: str, keep: bool, mount_src: str,
                     attach: bool) -> None:
    """
    Start the server.
    """
    # will print an error if not
    check_docker_running()

    info("Starting server...")
    info("Finding Docker daemon.")
    docker_client = docker.from_env()

    # check if name is allowed for docker volume, else exit
    check_config_name_allowed(ctx.name)

    # check that this server is not already running
    running_servers = docker_client.containers.list(
        filters={"label": f"{APPNAME}-type=server"})
    for server in running_servers:
        if server.name == f"{APPNAME}-{ctx.name}-{ctx.scope}-server":
            error(f"Server {Fore.RED}{ctx.name}{Style.RESET_ALL} "
                  "is already running")
            exit(1)

    # check that log directory exists - or create it
    ctx.log_dir.mkdir(parents=True, exist_ok=True)

    # Determine image-name. First we check if the option --image has been used.
    # Then we check if the image has been specified in the config file, and
    # finally we use the default settings from the package.
    if image is None:
        custom_images: dict = ctx.config.get('images')
        if custom_images:
            image = custom_images.get('server')
        if not image:
            image = f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_SERVER_IMAGE}"

    info(f"Pulling latest server image '{image}'.")
    try:
        pull_if_newer(docker.from_env(), image)
        # docker_client.images.pull(image)
    except Exception as e:
        warning(' ... Getting latest server image failed:')
        warning(f"     {e}")
    else:
        info(" ... success!")

    info("Creating mounts")
    config_file = "/mnt/config.yaml"
    mounts = [
        docker.types.Mount(
            config_file, str(ctx.config_file), type="bind"
        ), docker.types.Mount(
            "/mnt/log/", str(ctx.log_dir), type="bind"
        )
    ]

    if mount_src:
        mount_src = os.path.abspath(mount_src)
        mounts.append(docker.types.Mount("/vantage6", mount_src, type="bind"))
    # FIXME: code duplication with cli_server_import()
    # try to mount database
    uri = ctx.config['uri']
    url = make_url(uri)
    environment_vars = None

    # If host is None, we're dealing with a file-based DB, like SQLite
    if (url.host is None):
        db_path = url.database

        if not os.path.isabs(db_path):
            # We're dealing with a relative path here -> make it absolute
            db_path = ctx.data_dir / url.database

        basename = os.path.basename(db_path)
        dirname = os.path.dirname(db_path)
        os.makedirs(dirname, exist_ok=True)

        # we're mounting the entire folder that contains the database
        mounts.append(docker.types.Mount(
            "/mnt/database/", dirname, type="bind"
        ))

        environment_vars = {
            "VANTAGE6_DB_URI": f"sqlite:////mnt/database/{basename}",
            "VANTAGE6_CONFIG_NAME": ctx.config_file_name
        }

    else:
        warning(f"Database could not be transferred, make sure {url.host} "
                "is reachable from the Docker container")
        info("Consider using the docker-compose method to start a server")

    # Create a docker network for the server and other services like RabbitMQ
    # to reside in
    server_network_mgr = NetworkManager(
        network_name=f"{APPNAME}-{ctx.name}-{ctx.scope}-network"
    )
    server_network_mgr.create_network(is_internal=False)

    if start_rabbitmq or ctx.config.get('rabbitmq') and \
            ctx.config['rabbitmq'].get('start_with_server', False):
        # Note that ctx.data_dir has been created at this point, which is
        # required for putting some RabbitMQ configuration files inside
        info('Starting RabbitMQ container')
        _start_rabbitmq(ctx, rabbitmq_image, server_network_mgr)
    elif ctx.config.get('rabbitmq'):
        info("RabbitMQ is provided in the config file as external service. "
             "Assuming this service is up and running.")
    else:
        warning('Message queue disabled! This means that the vantage6 server '
                'cannot be scaled horizontally!')

    # start the UI if requested
    if start_ui or ctx.config.get('ui') and ctx.config['ui'].get('enabled'):
        _start_ui(docker_client, ctx, ui_port)

    # The `ip` and `port` refer here to the ip and port within the container.
    # So we do not really care that is it listening on all interfaces.
    internal_port = 5000
    cmd = (
        f'uwsgi --http :{internal_port} --gevent 1000 --http-websockets '
        '--master --callable app --disable-logging '
        '--wsgi-file /vantage6/vantage6-server/vantage6/server/wsgi.py '
        f'--pyargv {config_file}'
    )
    info(cmd)

    info("Run Docker container")
    port_ = str(port or ctx.config["port"] or 5000)
    container = docker_client.containers.run(
        image,
        command=cmd,
        mounts=mounts,
        detach=True,
        labels={
            f"{APPNAME}-type": "server",
            "name": ctx.config_file_name
        },
        environment=environment_vars,
        ports={f"{internal_port}/tcp": (ip, port_)},
        name=ctx.docker_container_name,
        auto_remove=not keep,
        tty=True,
        network=server_network_mgr.network_name
    )

    info(f"Success! container id = {container.id}")

    if attach:
        logs = container.attach(stream=True, logs=True, stdout=True)
        Thread(target=print_log_worker, args=(logs,), daemon=True).start()
        while True:
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                info("Closing log file. Keyboard Interrupt.")
                info("Note that your server is still running! Shut it down "
                     f"with {Fore.RED}v6 server stop{Style.RESET_ALL}")
                exit(0)


def _start_rabbitmq(ctx: ServerContext, rabbitmq_image: str,
                    network_mgr: NetworkManager) -> None:
    """
    Start the RabbitMQ container if it is not already running.

    Parameters
    ----------
    ctx : ServerContext
        Server context object
    rabbitmq_image : str
        RabbitMQ image to use
    network_mgr : NetworkManager
        Network manager object
    """
    rabbit_uri = ctx.config['rabbitmq'].get('uri')
    if not rabbit_uri:
        error("No RabbitMQ URI found in the configuration file! Please add"
              "a 'uri' key to the 'rabbitmq' section of the configuration.")
        exit(1)
    # kick off RabbitMQ container
    rabbit_mgr = RabbitMQManager(
        ctx=ctx, network_mgr=network_mgr, image=rabbitmq_image)
    rabbit_mgr.start()


def _start_ui(client: DockerClient, ctx: ServerContext, ui_port: int) -> None:
    """
    Start the UI container.

    Parameters
    ----------
    client : DockerClient
        Docker client
    ctx : ServerContext
        Server context object
    ui_port : int
        Port to expose the UI on
    """
    # if no port is specified, check if config contains a port
    ui_config = ctx.config.get('ui')
    if ui_config and not ui_port:
        ui_port = ui_config.get('port')

    # check if the port is valid
    # TODO make function to check if port is valid, and use in more places
    if not isinstance(ui_port, int) or not 0 < ui_port < 65536:
        warning(f"UI port '{ui_port}' is not valid! Using default port "
                f"{DEFAULT_UI_PORT}")
        ui_port = DEFAULT_UI_PORT

    # find image to use
    custom_images: dict = ctx.config.get('images')
    image = None
    if custom_images:
        image = custom_images.get('ui')
    if not image:
        image = f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_UI_IMAGE}"

    info(f"Pulling latest UI image '{image}'.")
    try:
        pull_if_newer(docker.from_env(), image)
        # docker_client.images.pull(image)
    except Exception as e:
        warning(' ... Getting latest node image failed:')
        warning(f"     {e}")
    else:
        info(" ... success!")

    # set environment variables
    env_vars = {
        "SERVER_URL": f"http://localhost:{ctx.config.get('port')}",
        "API_PATH": ctx.config.get("api_path"),
    }

    # stop the UI container if it is already running
    stop_ui(client, ctx)

    info(f'Starting User Interface at port {ui_port}')
    ui_container_name = f"{APPNAME}-{ctx.name}-{ctx.scope}-ui"
    client.containers.run(
        image,
        detach=True,
        labels={
            f"{APPNAME}-type": "ui",
            "name": ctx.config_file_name
        },
        ports={"80/tcp": (ctx.config.get('ip'), ui_port)},
        name=ui_container_name,
        environment=env_vars,
        tty=True,
    )



import click
from colorama import Fore, Style

from vantage6.common import error, info, check_config_writeable
from vantage6.cli.context import NodeContext
from vantage6.cli.globals import (
    DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
)
from vantage6.cli.configuration_wizard import configuration_wizard
from vantage6.cli.utils import check_config_name_allowed, prompt_config_name


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option('--system', 'system_folders', flag_value=True,
              help="Store this configuration in the system folders")
@click.option('--user', 'system_folders', flag_value=False, default=N_FOL,
              help="Store this configuration in the user folders. This is the "
                   "default")
def cli_node_new_configuration(name: str, system_folders: bool) -> None:
    """
    Create a new node configuration.

    Checks if the configuration already exists. If this is not the case
    a questionnaire is invoked to create a new configuration file.
    """
    name = prompt_config_name(name)
    # check if config name is allowed docker name
    check_config_name_allowed(name)

    # check that this config does not exist
    if NodeContext.config_exists(name, system_folders):
        error(f"Configuration {name} already exists!")
        exit(1)

    # Check that we can write in this folder
    if not check_config_writeable(system_folders):
        error("Cannot write configuration file. Exiting...")
        exit(1)

    # create config in ctx location
    flag = "--system" if system_folders else ""
    cfg_file = configuration_wizard("node", name, system_folders)
    info(f"New configuration created: {Fore.GREEN}{cfg_file}{Style.RESET_ALL}")
    info(f"You can start the node by running "
         f"{Fore.GREEN}v6 node start {flag}{Style.RESET_ALL}")



from pathlib import Path
import click

from vantage6.cli.utils import info
from vantage6.cli.dev.create import create_demo_network
from vantage6.cli.dev.start import start_demo_network
from vantage6.cli.dev.stop import stop_demo_network
from vantage6.cli.dev.remove import remove_demo_network
from vantage6.cli.utils import prompt_config_name, check_config_name_allowed
from vantage6.cli.test.feature_tester import cli_test_features


@click.command()
@click.option('-n', '--name', default=None, type=str,
              help="Name for your development setup")
@click.option('--server-url', type=str, default='http://host.docker.internal',
              help='Server URL to point to. If you are using Docker Desktop, '
              'the default http://host.docker.internal should not be changed.')
@click.option('-i', '--image', type=str, default=None,
              help='Server Docker image to use')
@click.option('--keep', type=bool, default=False,
              help='Keep the dev network after finishing the test')
@click.option('--extra-server-config', type=click.Path(exists=True),
              default=None, help='YAML File with additional server '
              'configuration. This will be appended to the server '
              'configuration file')
@click.option('--extra-node-config', type=click.Path('rb'), default=None,
              help='YAML File with additional node configuration. This will be'
              ' appended to each of the node configuration files')
@click.pass_context
def cli_test_integration(
    click_ctx: click.Context, name: str, server_url: str, image: str,
    keep: bool = False, extra_server_config: Path = None,
    extra_node_config: Path = None
) -> list[dict]:
    """
    Create dev network and run diagnostic checks on it.

    This is a full integration test of the vantage6 network. It will create
    a test server with some nodes using the `vdev` commands, and then run the
    v6-diagnostics algorithm to test all functionality.
    """
    # get name for the development setup - if not given - and check if it is
    # allowed
    name = prompt_config_name(name)
    check_config_name_allowed(name)

    # create server & node configurations and create test resources (
    # collaborations, organizations, etc)
    click_ctx.invoke(
        create_demo_network, name=name, num_nodes=3, server_url=server_url,
        server_port=5000, image=image, extra_server_config=extra_server_config,
        extra_node_config=extra_node_config
    )

    # start the server and nodes
    click_ctx.invoke(
        start_demo_network, name=name, system_folders=True, server_image=image,
        node_image=image
    )

    # run the diagnostic tests
    # TODO the username and password should be coordinated with the vdev
    # command - at present it spits out this username/password combination by
    # default but both should be defined in the same place
    # TODO VPN testing is always excluded - allow to include it with a flag
    # when vdev commands can handle extra config parameters
    diagnose_results = click_ctx.invoke(
        cli_test_features, host="http://localhost", port=5000, api_path='/api',
        username='org_1-admin', password='password', collaboration=1,
        organizations=None, all_nodes=True, online_only=False, no_vpn=True
    )

    # clean up the test resources
    click_ctx.invoke(stop_demo_network, name=name, system_folders=True)
    if not keep:
        click_ctx.invoke(remove_demo_network, name=name, system_folders=True)
    else:
        info("Keeping the demo network {name}. You can start it with `v6 dev "
             "start-demo-network`")

    return diagnose_results



"""
Utility functions for the CLI
"""
from __future__ import annotations

import re
import docker
import os
import questionary as q

from pathlib import Path

from vantage6.common import error, warning, info


def check_config_name_allowed(name: str) -> None:
    """
    Check if configuration name is allowed

    Parameters
    ----------
    name : str
        Name to be checked
    """
    if name.count(" ") > 0:
        name = name.replace(" ", "-")
        info(f"Replaced spaces from configuration name: {name}")
    if not re.match('^[a-zA-Z0-9_.-]+$', name):
        error(f"Name '{name}' is not allowed. Please use only the following "
              "characters: a-zA-Z0-9_.-")
        # FIXME: FM, 2023-01-03: I dont think this is a good side effect. This
        # should be handled by the caller.
        exit(1)


def check_if_docker_daemon_is_running(
        docker_client: docker.DockerClient) -> None:
    """
    Check if Docker daemon is running

    Parameters
    ----------
    docker_client : docker.DockerClient
        The docker client
    """
    try:
        docker_client.ping()
    except Exception:
        error("Docker socket can not be found. Make sure Docker is running.")
        exit(1)


def remove_file(file: str | Path, file_type: str) -> None:
    """
    Remove a file if it exists.

    Parameters
    ----------
    file : str
        absolute path to the file to be deleted
    file_type : str
        type of file, used for logging
    """
    if os.path.isfile(file):
        info(f"Removing {file_type} file: {file}")
        try:
            os.remove(file)
        except Exception as e:
            error(f"Could not delete file: {file}")
            error(e)
    else:
        warning(f"Could not remove {file_type} file: {file} does not exist")


def prompt_config_name(name: str | None) -> None:
    """
    Get a new configuration name from the user, or simply return the name if
    it is not None.

    Parameters
    ----------
    name : str
        Name to be checked

    Returns
    -------
    str
        The name of the configuration
    """
    if not name:
        name = q.text("Please enter a configuration-name:").ask()
        if name.count(" ") > 0:
            name = name.replace(" ", "-")
            info(f"Replaced spaces from configuration name: {name}")
    return name



import click
import questionary as q
import docker

from colorama import Fore, Style

from vantage6.common import warning, error, info
from vantage6.common.globals import APPNAME
from vantage6.common.docker.addons import check_docker_running
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL

from vantage6.cli.node.common import find_running_node_names


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option('--system', 'system_folders', flag_value=True,
              help="Search for configuration in system folders instead of "
                   "user folders")
@click.option('--user', 'system_folders', flag_value=False, default=N_FOL,
              help="Search for configuration in the user folders instead of "
                   "system folders. This is the default.")
@click.option('--all', 'all_nodes', flag_value=True,
              help="Stop all running nodes")
@click.option('--force', 'force', flag_value=True,
              help="Kill nodes instantly; don't wait for them to shut down")
def cli_node_stop(name: str, system_folders: bool, all_nodes: bool,
                  force: bool) -> None:
    """
    Stop one or all running nodes.
    """
    check_docker_running()
    client = docker.from_env()

    running_node_names = find_running_node_names(client)

    if not running_node_names:
        warning("No nodes are currently running.")
        return

    if force:
        warning('Forcing the node to stop will not terminate helper '
                'containers, neither will it remove routing rules made on the '
                'host!')

    if all_nodes:
        for name in running_node_names:
            container = client.containers.get(name)
            if force:
                container.kill()
            else:
                container.stop()
            info(f"Stopped the {Fore.GREEN}{name}{Style.RESET_ALL} Node.")
    else:
        if not name:
            name = q.select("Select the node you wish to stop:",
                            choices=running_node_names).ask()
        else:

            post_fix = "system" if system_folders else "user"
            name = f"{APPNAME}-{name}-{post_fix}"

        if name in running_node_names:
            container = client.containers.get(name)
            # Stop the container. Using stop() gives the container 10s to exit
            # itself, if not then it will be killed
            if force:
                container.kill()
            else:
                container.stop()
            info(f"Stopped the {Fore.GREEN}{name}{Style.RESET_ALL} Node.")
        else:
            error(f"{Fore.RED}{name}{Style.RESET_ALL} is not running?")



import click
import questionary as q
import docker

from vantage6.common import error
from vantage6.common.docker.addons import check_docker_running
from vantage6.common.globals import APPNAME
from vantage6.cli.globals import DEFAULT_SERVER_SYSTEM_FOLDERS
from vantage6.cli import __version__


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option('--system', 'system_folders', flag_value=True)
@click.option('--user', 'system_folders', flag_value=False,
              default=DEFAULT_SERVER_SYSTEM_FOLDERS)
def cli_server_version(name: str, system_folders: bool) -> None:
    """
    Print the version of the vantage6 server.
    """
    check_docker_running()
    client = docker.from_env()

    running_servers = client.containers.list(
        filters={"label": f"{APPNAME}-type=server"})
    running_server_names = [server.name for server in running_servers]

    if not name:
        if not running_server_names:
            error("No servers are running! You can only check the version for "
                  "servers that are running")
            exit(1)
        name = q.select("Select the server you wish to inspect:",
                        choices=running_server_names).ask()
    else:
        post_fix = "system" if system_folders else "user"
        name = f"{APPNAME}-{name}-{post_fix}"

    if name in running_server_names:
        container = client.containers.get(name)
        version = container.exec_run(cmd='vserver-local version',
                                     stdout=True)
        click.echo({"server": version.output.decode('utf-8'),
                    "cli": __version__})
    else:
        error(f"Server {name} is not running! Cannot provide version...")



import sys
from pathlib import Path
from threading import Thread
import time
import os.path

import click
import questionary as q
import docker

from colorama import Fore, Style

from vantage6.common import warning, error, info, debug, get_database_config
from vantage6.common.globals import (
    APPNAME,
    DEFAULT_DOCKER_REGISTRY,
    DEFAULT_NODE_IMAGE,
    DEFAULT_NODE_IMAGE_WO_TAG,
)
from vantage6.common.docker.addons import (
    pull_if_newer,
    remove_container_if_exists,
    check_docker_running,
)

from vantage6.cli.context import NodeContext
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli.configuration_wizard import (
    configuration_wizard,
    select_configuration_questionaire,
)
from vantage6.cli.utils import check_config_name_allowed
from vantage6.cli import __version__
from vantage6.cli.node.common import print_log_worker, create_client


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option(
    "-c", "--config", default=None, help="Path to configuration-file; overrides NAME"
)
@click.option(
    "--system",
    "system_folders",
    flag_value=True,
    help="Search for the configuration in the system folders",
)
@click.option(
    "--user",
    "system_folders",
    flag_value=False,
    default=N_FOL,
    help="Search for the configuration in the user folders. This is " "the default",
)
@click.option("-i", "--image", default=None, help="Node Docker image to use")
@click.option(
    "--keep/--auto-remove",
    default=False,
    help="Keep node container after finishing. Useful for debugging",
)
@click.option(
    "--force-db-mount",
    is_flag=True,
    help="Always mount node databases; skip the check if they are " "existing files.",
)
@click.option(
    "--attach/--detach",
    default=False,
    help="Show node logs on the current console after starting the " "node",
)
@click.option(
    "--mount-src",
    default="",
    help="Override vantage6 source code in container with the source"
    " code in this path",
)
def cli_node_start(
    name: str,
    config: str,
    system_folders: bool,
    image: str,
    keep: bool,
    mount_src: str,
    attach: bool,
    force_db_mount: bool,
) -> None:
    """
    Start the node.
    """
    check_docker_running()
    info("Starting node...")
    info("Finding Docker daemon")
    docker_client = docker.from_env()
    NodeContext.LOGGING_ENABLED = False
    if config:
        name = Path(config).stem
        ctx = NodeContext(name, system_folders, config)

    else:
        # in case no name is supplied, ask the user to select one
        if not name:
            name = select_configuration_questionaire("node", system_folders)

        # check that config exists, if not a questionaire will be invoked
        if not NodeContext.config_exists(name, system_folders):
            warning(f"Configuration {Fore.RED}{name}{Style.RESET_ALL} does not exist.")

            if q.confirm("Create this configuration now?").ask():
                configuration_wizard("node", name, system_folders)

            else:
                error("Config file couldn't be loaded")
                sys.exit(0)

        ctx = NodeContext(name, system_folders)

    # check if config name is allowed docker name, else exit
    check_config_name_allowed(ctx.name)

    # check that this node is not already running
    running_nodes = docker_client.containers.list(
        filters={"label": f"{APPNAME}-type=node"}
    )

    suffix = "system" if system_folders else "user"
    for node in running_nodes:
        if node.name == f"{APPNAME}-{name}-{suffix}":
            error(f"Node {Fore.RED}{name}{Style.RESET_ALL} is already running")
            exit(1)

    # make sure the (host)-task and -log dir exists
    info("Checking that data and log dirs exist")
    ctx.data_dir.mkdir(parents=True, exist_ok=True)
    ctx.log_dir.mkdir(parents=True, exist_ok=True)

    # Determine image-name. First we check if the option --image has been used.
    # Then we check if the image has been specified in the config file, and
    # finally we use the default settings from the package.
    if not image:
        custom_images: dict = ctx.config.get("images")
        if custom_images:
            image = custom_images.get("node")
        else:
            # if no custom image is specified, find the server version and use
            # the latest images from that minor version
            client = create_client(ctx)
            major_minor = None
            try:
                # try to get server version, skip if can't get a connection
                version = client.util.get_server_version(attempts_on_timeout=3)[
                    "version"
                ]
                major_minor = ".".join(version.split(".")[:2])
                image = (
                    f"{DEFAULT_DOCKER_REGISTRY}/"
                    f"{DEFAULT_NODE_IMAGE_WO_TAG}"
                    f":{major_minor}"
                )
            except Exception:
                warning(
                    "Could not determine server version. Using default " "node image"
                )

            if major_minor and not __version__.startswith(major_minor):
                warning(
                    "Version mismatch between CLI and server/node. CLI is "
                    f"running on version {__version__}, while node and server "
                    f"are on version {major_minor}. This might cause "
                    f"unexpected issues; changing to {major_minor}.<latest> "
                    "is recommended."
                )

        # fail safe, in case no custom image is specified and we can't get the
        # server version
        if not image:
            image = f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_NODE_IMAGE}"

    info(f"Pulling latest node image '{image}'")
    try:
        # docker_client.images.pull(image)
        pull_if_newer(docker.from_env(), image)

    except Exception as e:
        warning(" ... Getting latest node image failed:")
        warning(f"     {e}")
    else:
        info(" ... success!")

    info("Creating Docker data volume")

    data_volume = docker_client.volumes.create(ctx.docker_volume_name)
    vpn_volume = docker_client.volumes.create(ctx.docker_vpn_volume_name)
    ssh_volume = docker_client.volumes.create(ctx.docker_ssh_volume_name)
    squid_volume = docker_client.volumes.create(ctx.docker_squid_volume_name)

    info("Creating file & folder mounts")
    # FIXME: should obtain mount points from DockerNodeContext
    mounts = [
        # (target, source)
        (f"/mnt/log/{name}", str(ctx.log_dir)),
        ("/mnt/data", data_volume.name),
        ("/mnt/vpn", vpn_volume.name),
        ("/mnt/ssh", ssh_volume.name),
        ("/mnt/squid", squid_volume.name),
        ("/mnt/config", str(ctx.config_dir)),
        ("/var/run/docker.sock", "/var/run/docker.sock"),
    ]

    if mount_src:
        # If mount_src is a relative path, docker will consider it a volume.
        mount_src = os.path.abspath(mount_src)
        mounts.append(("/vantage6", mount_src))

    # FIXME: Code duplication: Node.__init__() (vantage6/node/__init__.py)
    #   uses a lot of the same logic. Suggest moving this to
    #   ctx.get_private_key()
    filename = ctx.config.get("encryption", {}).get("private_key")
    # filename may be set to an empty string
    if not filename:
        filename = "private_key.pem"

    # Location may be overridden by the environment
    filename = os.environ.get("PRIVATE_KEY", filename)

    # If ctx.get_data_file() receives an absolute path, it is returned as-is
    fullpath = Path(ctx.get_data_file(filename))
    if fullpath:
        if Path(fullpath).exists():
            mounts.append(("/mnt/private_key.pem", str(fullpath)))
        else:
            warning(f"private key file provided {fullpath}, " "but does not exists")

    # Mount private keys for ssh tunnels
    ssh_tunnels = ctx.config.get("ssh-tunnels", [])
    for ssh_tunnel in ssh_tunnels:
        hostname = ssh_tunnel.get("hostname")
        key_path = ssh_tunnel.get("ssh", {}).get("identity", {}).get("key")
        if not key_path:
            error(
                f"SSH tunnel identity {Fore.RED}{hostname}{Style.RESET_ALL} "
                "key not provided. Continuing to start without this tunnel."
            )
        key_path = Path(key_path)
        if not key_path.exists():
            error(
                f"SSH tunnel identity {Fore.RED}{hostname}{Style.RESET_ALL} "
                "key does not exist. Continuing to start without this "
                "tunnel."
            )

        info(f"  Mounting private key for {hostname} at {key_path}")

        # we remove the .tmp in the container, this is because the file is
        # mounted in a volume mount point. Somehow the file is than empty in
        # the volume but not for the node instance. By removing the .tmp we
        # make sure that the file is not empty in the volume.
        mounts.append((f"/mnt/ssh/{hostname}.pem.tmp", str(key_path)))

    env = {
        "DATA_VOLUME_NAME": data_volume.name,
        "VPN_VOLUME_NAME": vpn_volume.name,
        "PRIVATE_KEY": "/mnt/private_key.pem",
    }

    # only mount the DB if it is a file
    info("Setting up databases")
    db_labels = [db["label"] for db in ctx.databases]
    for label in db_labels:

        # check that label contains only valid characters
        if not label.isidentifier():
            error(f"Database label {Fore.RED}{label}{Style.RESET_ALL} contains"
                  " invalid characters. Only letters, numbers, and underscores"
                  " are allowed, and it cannot start with a number.")
            exit(1)

        db_config = get_database_config(ctx.databases, label)
        uri = db_config["uri"]
        db_type = db_config["type"]

        info(
            f"  Processing {Fore.GREEN}{db_type}{Style.RESET_ALL} database "
            f"{Fore.GREEN}{label}:{uri}{Style.RESET_ALL}"
        )
        label_capitals = label.upper()

        try:
            file_based = Path(uri).exists()
        except Exception:
            # If the database uri cannot be parsed, it is definitely not a
            # file. In case of http servers or sql servers, checking the path
            # of the the uri will lead to an OS-dependent error, which is why
            # we catch all exceptions here.
            file_based = False

        if not file_based and not force_db_mount:
            debug("  - non file-based database added")
            env[f"{label_capitals}_DATABASE_URI"] = uri
        else:
            debug("  - file-based database added")
            suffix = Path(uri).suffix
            env[f"{label_capitals}_DATABASE_URI"] = f"{label}{suffix}"
            mounts.append((f"/mnt/{label}{suffix}", str(uri)))

    system_folders_option = "--system" if system_folders else "--user"
    cmd = (
        f"vnode-local start -c /mnt/config/{name}.yaml -n {name} "
        f" --dockerized {system_folders_option}"
    )

    info("Running Docker container")
    volumes = []
    for mount in mounts:
        volumes.append(f"{mount[1]}:{mount[0]}")

    remove_container_if_exists(
        docker_client=docker_client, name=ctx.docker_container_name
    )

    container = docker_client.containers.run(
        image,
        command=cmd,
        volumes=volumes,
        detach=True,
        labels={
            f"{APPNAME}-type": "node",
            "system": str(system_folders),
            "name": ctx.config_file_name,
        },
        environment=env,
        name=ctx.docker_container_name,
        auto_remove=not keep,
        tty=True,
    )

    info("Node container was successfully started!")

    if attach:
        logs = container.attach(stream=True, logs=True)
        Thread(target=print_log_worker, args=(logs,), daemon=True).start()
        while True:
            try:
                time.sleep(1)
            except KeyboardInterrupt:
                info("Closing log file. Keyboard Interrupt.")
                info(
                    "Note that your node is still running! Shut it down with "
                    f"'{Fore.RED}v6 node stop{Style.RESET_ALL}'"
                )
                exit(0)
    else:
        info(f"To see the logs, run: {Fore.GREEN}v6 node attach --name "
             f"{ctx.name}{Style.RESET_ALL}")



import click
import questionary as q
import docker

from vantage6.common import error
from vantage6.common.globals import APPNAME
from vantage6.common.docker.addons import check_docker_running
from vantage6.cli.globals import DEFAULT_NODE_SYSTEM_FOLDERS as N_FOL
from vantage6.cli import __version__
from vantage6.cli.node.common import find_running_node_names


@click.command()
@click.option("-n", "--name", default=None, help="Configuration name")
@click.option('--system', 'system_folders', flag_value=True,
              help="Search for configuration in system folders rather than "
                   "user folders")
@click.option('--user', 'system_folders', flag_value=False, default=N_FOL,
              help="Search for configuration in user folders rather than "
                   "system folders. This is the default")
def cli_node_version(name: str, system_folders: bool) -> None:
    """
    Returns current version of a vantage6 node.
    """
    check_docker_running()
    client = docker.from_env()

    running_node_names = find_running_node_names(client)

    if not name:
        if not running_node_names:
            error("No nodes are running! You can only check the version for "
                  "nodes that are running")
            exit(1)
        name = q.select("Select the node you wish to inspect:",
                        choices=running_node_names).ask()
    else:
        post_fix = "system" if system_folders else "user"
        name = f"{APPNAME}-{name}-{post_fix}"

    if name in running_node_names:
        container = client.containers.get(name)
        version = container.exec_run(cmd='vnode-local version', stdout=True)
        click.echo(
            {"node": version.output.decode('utf-8'), "cli": __version__})
    else:
        error(f"Node {name} is not running! Cannot provide version...")



import codecs
import os

from os import path
from codecs import open
from setuptools import setup, find_namespace_packages
from pathlib import Path

# get current directory
here = Path(path.abspath(path.dirname(__file__)))
parent_dir = here.parent.absolute()

# get the long description from the README file
with codecs.open(path.join(parent_dir, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Read the API version from disk. This file should be located in the package
# folder, since it's also used to set the pkg.__version__ variable.
version_path = os.path.join(here, 'vantage6', 'cli', '_version.py')
version_ns = {
    '__file__': version_path
}
with open(version_path) as f:
    exec(f.read(), {}, version_ns)


# setup the package
setup(
    name='vantage6',
    version=version_ns['__version__'],
    description='vantage6 command line interface',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/vantage6/vantage6',
    packages=find_namespace_packages(),
    python_requires='>=3.10',
    install_requires=[
        'click==8.1.3',
        'colorama==0.4.6',
        'copier==8.3.0',
        'docker==6.1.2',
        'ipython==8.10.0',
        'jinja2==3.1.3',
        'questionary==1.10.0',
        'rich==13.5.2',
        'schema==0.7.5',
        'SQLAlchemy==1.4.46',
        f'vantage6-common == {version_ns["__version__"]}',
        f'vantage6-client == {version_ns["__version__"]}',
    ],
    extras_require={
        'dev': [
            'coverage==6.4.4'
        ]
    },
    package_data={
        'vantage6.cli': [
            '__build__',
            'rabbitmq/rabbitmq.config',
            'template/node_config.j2',
            'template/server_config.j2',
            'template/server_import_config.j2',
        ],
    },
    entry_points={
        'console_scripts': [
            'v6=vantage6.cli.cli:cli_complete',
            # TODO v5+ remove vnode/vserver/vdev aliases (?)
            'vnode=vantage6.cli.cli:cli_node',
            'vserver=vantage6.cli.cli:cli_server',
            'vdev=vantage6.cli.cli:cli_dev',
        ]
    }
)



import click

from vantage6.cli.server.attach import cli_server_attach
from vantage6.cli.server.files import cli_server_files
from vantage6.cli.server.import_ import cli_server_import
from vantage6.cli.server.list import cli_server_configuration_list
from vantage6.cli.server.new import cli_server_new
from vantage6.cli.server.remove import cli_server_remove
from vantage6.cli.server.shell import cli_server_shell
from vantage6.cli.server.start import cli_server_start
from vantage6.cli.server.stop import cli_server_stop
from vantage6.cli.server.version import cli_server_version
from vantage6.cli.node.attach import cli_node_attach
from vantage6.cli.node.clean import cli_node_clean
from vantage6.cli.node.create_private_key import cli_node_create_private_key
from vantage6.cli.node.files import cli_node_files
from vantage6.cli.node.list import cli_node_list
from vantage6.cli.node.new import cli_node_new_configuration
from vantage6.cli.node.remove import cli_node_remove
from vantage6.cli.node.set_api_key import cli_node_set_api_key
from vantage6.cli.node.start import cli_node_start
from vantage6.cli.node.stop import cli_node_stop
from vantage6.cli.node.version import cli_node_version
from vantage6.cli.dev.create import create_demo_network
from vantage6.cli.dev.remove import remove_demo_network
from vantage6.cli.dev.start import start_demo_network
from vantage6.cli.dev.stop import stop_demo_network
from vantage6.cli.algorithm.create import cli_algorithm_create
from vantage6.cli.algorithm.update import cli_algorithm_update
from vantage6.cli.test.feature_tester import cli_test_features
from vantage6.cli.test.integration_test import cli_test_integration


# Define the server group
@click.group(name='server')
def cli_server() -> None:
    """
    Manage your vantage6 server instances.
    """


# Define the commands for the server group
cli_server.add_command(cli_server_attach, name='attach')
cli_server.add_command(cli_server_files, name='files')
cli_server.add_command(cli_server_import, name='import')
cli_server.add_command(cli_server_configuration_list, name='list')
cli_server.add_command(cli_server_new, name='new')
cli_server.add_command(cli_server_remove, name='remove')
cli_server.add_command(cli_server_shell, name='shell')
cli_server.add_command(cli_server_start, name='start')
cli_server.add_command(cli_server_stop, name='stop')
cli_server.add_command(cli_server_version, name='version')


# Define the node group
@click.group(name="node")
def cli_node() -> None:
    """
    Manage your vantage6 node instances.
    """


# Define the commands for the node group
cli_node.add_command(cli_node_attach, name="attach")
cli_node.add_command(cli_node_clean, name="clean")
cli_node.add_command(cli_node_create_private_key, name="create-private-key")
cli_node.add_command(cli_node_files, name="files")
cli_node.add_command(cli_node_list, name="list")
cli_node.add_command(cli_node_new_configuration, name="new")
cli_node.add_command(cli_node_remove, name="remove")
cli_node.add_command(cli_node_set_api_key, name="set-api-key")
cli_node.add_command(cli_node_start, name="start")
cli_node.add_command(cli_node_stop, name="stop")
cli_node.add_command(cli_node_version, name="version")


# Define the dev group
@click.group(name="dev")
def cli_dev() -> None:
    """
    Quickly manage a test network with a server and several nodes.

    These commands are helpful for local testing of your vantage6 environment.
    """


# Define the commands for the dev group
cli_dev.add_command(create_demo_network, name="create-demo-network")
cli_dev.add_command(remove_demo_network, name="remove-demo-network")
cli_dev.add_command(start_demo_network, name="start-demo-network")
cli_dev.add_command(stop_demo_network, name="stop-demo-network")


# Define the algorithm group
@click.group(name="algorithm")
def cli_algorithm() -> None:
    """
    Manage your vantage6 algorithms.
    """


# Define the commands for the algorithm group
cli_algorithm.add_command(cli_algorithm_create, name="create")
cli_algorithm.add_command(cli_algorithm_update, name="update")


# Define the test group
@click.group(name="test")
def cli_test() -> None:
    """
    Execute tests on your vantage6 infrastructure.
    """


# Define the commands for the test group
cli_test.add_command(cli_test_features, name="feature-test")
cli_test.add_command(cli_test_integration, name="integration-test")


# Define the overall group
@click.group(name='cli')
def cli_complete() -> None:
    """
    The `v6` command line interface allows you to manage your vantage6
    infrastructure.

    It provides a number of subcommands to help you with this task.
    """


# Add the subcommands to the overall group
cli_complete.add_command(cli_node)
cli_complete.add_command(cli_server)
cli_complete.add_command(cli_dev)
cli_complete.add_command(cli_algorithm)
cli_complete.add_command(cli_test)
